'''
Modified WikiLinks Extension for Python-Markdown
======================================
Added ability to set custom pattern of markdown inline syntax via "pattern" option. Also label can be changed via second return value.
======================================

Converts [[WikiLinks]] to relative links.

See <https://Python-Markdown.github.io/extensions/wikilinks>
for documentation.

Original code Copyright [Waylan Limberg](http://achinghead.com/).

All changes Copyright The Python Markdown Project

License: [BSD](https://opensource.org/licenses/bsd-license.php)

'''

from markdown.extensions import Extension
from markdown.inlinepatterns import InlineProcessor
from markdown.extensions.wikilinks import build_url
import xml.etree.ElementTree as etree


class InternalLinkExtension(Extension):

	def __init__(self, **kwargs):
		self.config = {
			'base_url'  : ['/', 'String to append to beginning or URL.'],
			'end_url'   : ['/', 'String to append to end of URL.'],
			'html_class': ['wikilink', 'CSS hook. Leave blank for none.'],
			'build_url' : [build_url, 'Callable formats URL and label.'],
			'pattern'   : [r'\[\[([\w0-9_ -]+)\]\]', 'Regular expression pattern for markdown inline syntax.'],
		}

		super().__init__(**kwargs)

	def extendMarkdown(self, md):
		self.md = md

		# append to end of inline patterns
		INTERNALLINK_RE = self.config['pattern'][0]
		internallinkPattern = InternalLinksInlineProcessor(INTERNALLINK_RE, self.getConfigs())
		internallinkPattern.md = md
		md.inlinePatterns.register(internallinkPattern, 'internallink', 75)


class InternalLinksInlineProcessor(InlineProcessor):
	def __init__(self, pattern, config):
		super().__init__(pattern)
		self.config = config

	def handleMatch(self, m, data):
		if m.group(1).strip():
			base_url, end_url, html_class = self._getMeta()
			label = m.group(1).strip()
			url, label = self.config['build_url'](label, base_url, end_url)
			a = etree.Element('a')
			a.text = label
			a.set('href', url)
			if html_class:
				a.set('class', html_class)
		else:
			a = ''
		return a, m.start(0), m.end(0)

	def _getMeta(self):
		""" Return meta data or config data. """
		base_url = self.config['base_url']
		end_url = self.config['end_url']
		html_class = self.config['html_class']
		if hasattr(self.md, 'Meta'):
			if 'wiki_base_url' in self.md.Meta:
				base_url = self.md.Meta['wiki_base_url'][0]
			if 'wiki_end_url' in self.md.Meta:
				end_url = self.md.Meta['wiki_end_url'][0]
			if 'wiki_html_class' in self.md.Meta:
				html_class = self.md.Meta['wiki_html_class'][0]
		return base_url, end_url, html_class


def makeExtension(**kwargs):  # pragma: no cover
	return InternalLinkExtension(**kwargs)



"""Simple HTTP 1.1 web server."""
import mimetypes
import re
import socket
import urllib
from pathlib import Path
from typing import Optional, Callable, Tuple
from urllib.parse import urlparse


def parse_request(text: str) -> Optional[str]:
	"""
	Parse HTTP request and return requested path or None.
	"""
	if (match := re.match(r'\s*GET\s+([^\s]+)', text)) is not None:
		url = urlparse(match.group(1))
		if len(url.scheme) != 0 and url.scheme != 'http':
			return None
		elif len(url.path) == 0:
			return './'
		else:
			return '.' + urllib.parse.unquote(url.path)
	else:
		return None


def generate_response(body: bytes, type: str = None) -> bytes:
	"""
	Generate HTTP 1.0 response with 200 status code, Content-Length, Content-Type and Connection: close headers.
	"""
	charset = '; charset=utf-8' if type.startswith('text') else ''
	return ('\r\n'.join([
		'HTTP/1.1 200 OK',
		f'Content-Length: {len(body)}',
		f'Content-Type: {type or "application/octet-stream"}{charset}',
		'Connection: close',
	]) + '\r\n\r\n').encode('utf-8') + body


def default_routing_callback(requested_path: Path) -> Optional[Path]:
	"""
	Default callback for requested path routing that maps direcotry paths to directory/index.html ones.
	"""
	if '.' not in requested_path.name:
		return requested_path / 'index.html'
	else:
		return requested_path


def default_response_callback(requested_path: Path) -> Optional[Tuple[bytes, str]]:
	"""
	Default callback for response that return specified file content.
	"""
	if requested_path.exists() and requested_path.is_file():
		return requested_path.read_bytes(), mimetypes.guess_type(requested_path, strict=False)[0]


def serve(directory: Path = Path.cwd(), interface: str = '0.0.0.0', port: int = 80, routing_callback: Callable[[Path], Optional[Path]] = default_routing_callback, response_callback: Callable[[Path], Optional[Tuple[bytes, str]]] = default_response_callback):
	"""
	Listen forever.

	:param directory: directory to server as web root.
	:param interface: Interface IP v4 address or resolvable name (like "127.0.0.1" or "localhost") on which to serve. Use "0.0.0.0" for all connected interfaces.
	:param port: Port on which to serve.
	:param routing_callback: routing callback that must return requested path or None.
	:param response_callback:  response body generation callback that must return bytes and MIME type or None.
	"""
	print('Starting to serve', directory, f'at http://{interface or "localhost"}:{port}')
	with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server:
		server.bind((interface, port))
		server.listen(999)
		while True:
			client, addr = server.accept()
			print('Connected by', addr)
			request = client.recv(2048)  # according to https://stackoverflow.com/a/417184 maximum url length is up to 2000 characters so 2048 bytes buffer size must be enough ro receive main path header
			requested_path = parse_request(request.decode('utf-8'))
			print('\tRequested path ', requested_path)
			requested_path = routing_callback(directory / Path(requested_path))
			print('\tRouted path ', requested_path)
			if requested_path is not None:
				if requested_path.is_relative_to(directory):
					response = response_callback(requested_path)
					if response is not None:
						client.sendall(generate_response(*response))
						print('\tSent data to client')
			print('\tDisconnecting client\r\n')
			client.close()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['typer[all] == 0.3.2',
 'Markdown == 3.3.4',
 'Jinja2 == 2.11.3',
 'chardet == 4.0.0',
 'python-markdown-math == 0.8',
 'attributedict == 0.3.0']

entry_points = \
{'console_scripts': ['simplewiki = wikiserver:main']}

setup(name='pysimplewiki',
      version='0.1.2',
      description='Simple Wiki - это вики движок, отображающий файловую структуру в веб-энциклопедию; поддерживает статьи в форматах .html, .md. Simple Wiki написана на языке python, поддерживает гибкое расширение и стилизацию.',
      author='baterflyrity',
      author_email='baterflyrity@yandex.ru',
      url='https://github.com/baterflyrity/pysimplewiki',
      py_modules=['wikiserver'],
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='~=3.8',
     )
